<?php

/******************** ABOUT ******************************
MySQL2AppGini, m2a.php, a script to import a MySQL/MariaDB
schema to AXP format for use with AppGini.
(C) Copyright 2022, BigProf Software.
https://bigprof.com/appgini/m2a
**********************************************************/

new M2A;

class M2A {
	private $dbCredentials;
	private $link;
	private $axp;
	private $tables;
	private $error;

	public function __construct() {
		/******************** workflow ***************************/
		$this->dbCredentials = $this->getDbCredentials();
		$this->link = $this->connectToDb();

		$this->tables = $this->getAppTableNames();
		$this->axp = $this->importDbToAxp();

		$this->saveAndDownloadAxp();
		/******************** end of workflow ********************/			
	}

	private function importDbToAxp() {
		$axp = "<database>";
		$axp .= "<databaseName>{$this->dbCredentials['dbName']}</databaseName>";
		$axp .= $this->getAppEncodingXml();
		$axp .= $this->getAppTablesXml();
		$axp .= "</database>";

		return $axp;
	}

	private function getAppTablesXml() {
		$axp = '';

		foreach($this->tables as $tn) {
			$axp .= implode('', [
				'<table>',
				"<name>$tn</name>",
				'<caption><![CDATA[' . $this->caption($tn) . ']]></caption>',
				'<allowInsert>False</allowInsert>',
				'<allowUpdate>False</allowUpdate>',
				'<allowCSV>True</allowCSV>',
				'<allowDelete>False</allowDelete>',
				'<allowDeleteOfParents>False</allowDeleteOfParents>',
				'<allowFilters>True</allowFilters>',
				'<filterFirst>False</filterFirst>',
				'<allowPrint>True</allowPrint>',
				'<allowMassDelete>False</allowMassDelete>',
				'<homepageShowCount>False</homepageShowCount>',
				'<allowDVPrint>True</allowDVPrint>',
				'<allowSavingFilters>False</allowSavingFilters>',
				'<allowSelection>True</allowSelection>',
				'<hideSaveNewWhenEditing>False</hideSaveNewWhenEditing>',
				'<allowSorting>True</allowSorting>',
				'<description><![CDATA[Table imported using MySQL2AppGini on ' . date('Y-m-d') . ']]></description>',
				'<detailViewLabel><![CDATA[Detail View]]></detailViewLabel>',
				'<hideTableView>True</hideTableView>',
				'<quickSearch>1</quickSearch>',
				'<recordsPerPage>10</recordsPerPage>',
				'<redirectAfterInsert><![CDATA[]]></redirectAfterInsert>',
				'<filterCount>10</filterCount>',
				'<lstChildrenLinks></lstChildrenLinks>',
				'<separateDV>True</separateDV>',
				'<hideLinkHomepage>False</hideLinkHomepage>',
				'<hideLinkNavMenu>False</hideLinkNavMenu>',
				'<defaultSortField>0</defaultSortField>',
				'<defaultSortDirection>asc</defaultSortDirection>',
				'<childrenLinksTV><![CDATA[]]></childrenLinksTV>',
				'<parentChildrenSettings><![CDATA[]]></parentChildrenSettings>',
				'<tableIcon></tableIcon>',
				'<tablePaginationAlignment>0</tablePaginationAlignment>',
				'<enableDVABFloating>True</enableDVABFloating>',
				'<groupID>0</groupID>',
				'<allowHomepageAddNew>False</allowHomepageAddNew>',
				'<TVTemplateID>horizontal</TVTemplateID>',
				'<TVTemplateHasTitle>False</TVTemplateHasTitle>',
				'<TVTemplateHideCaptions>False</TVTemplateHideCaptions>',
				'<TVTemplateImageWidth>25%</TVTemplateImageWidth>',
				'<TVClasses></TVClasses>',
				'<DVClasses></DVClasses>',
				'<technicalDocumentation><![CDATA[]]></technicalDocumentation>',
				'<autofocusField>-1</autofocusField>',
				$this->getTableFieldsXml($tn),
				'</table>',
			]);
		}

		return $axp;
	}

	private function getAppEncodingXml() {
		$dbName = mysqli_real_escape_string($this->link, $this->dbCredentials['dbName']);
		$res = @mysqli_query(
			$this->link, 
			"SELECT DEFAULT_CHARACTER_SET_NAME
			FROM `information_schema`.`SCHEMATA`
			WHERE `SCHEMA_NAME`='$dbName'"
		);
		if(!$res || !($row = mysqli_fetch_row($res))) return '';

		$appEncoding = [
			'latin1' => 'iso-8859-1',
			'utf8' => 'UTF-8',
			'cp1256' => 'Windows-1256',
			'latin7' => 'iso-8869-4',
			'cp1257' => 'Windows-1257',
			'latin2' => 'iso-8859-2',
			'cp1250' => 'Windows-1250',
			'gb2312' => 'gb2312',
			'gbk' => 'HZ-GB-2312',
			'big5' => 'big5',
			'koi8r' => 'koi8-r',
			'koi8u' => 'koi8-u',
			'cp1251' => 'Windows-1251',
			'greek' => 'iso-8859-7',
			'greek' => 'Windows-1253',
			'ujis' => 'x-euc-jp',
			'eucjpms' => 'iso-2022-jp',
			'sjis' => 'shift_jis',
			'euckr' => 'ISO-2022-KR',
			'euckr' => 'ks_c_5601',
			'hebrew' => 'dos-862',
			'hebrew' => 'Windows-1255',
			'hebrew' => 'iso-8859-8',
			'tis620' => 'Windows-874',
			'latin5' => 'iso-8859-9',
			'latin5' => 'Windows-1254',
			'utf8' => 'Windows-1258',
			'utf8mb4' => 'UTF-8',
		];

		return "<charEncoding>{$appEncoding[$row[0]]}</charEncoding>" .
			"<noMySQLEncoding>False</noMySQLEncoding>" .
			($row[0] == 'utf8mb4' ? '<charEncodingMySQLMB4>True</charEncodingMySQLMB4>' : '');
	}

	private function setError($msg) {
		$this->error = $msg;
		return false;
	}

	private function connectToDb() {
		$link = @mysqli_connect(
			$this->dbCredentials['dbServer'],
			$this->dbCredentials['dbUser'],
			$this->dbCredentials['dbPass'],
			$this->dbCredentials['dbName']
		);

		if(!$link)
			$this->askForCredentialsAndExit([
				'error' => 'Error while connecting to the database server: ' . mysqli_connect_error()
			]);

		return $link;
	}

	private function failAndDie() {
		ob_start(); ?>
		<div class="danger">
			<?php echo $this->error; ?>
			<button type="button" onclick="history.go(-1);">&#9664; Try again</button>
		</div>
		<?php
		die(
			ob_get_clean() . 
			$this->copyrightFooter() . 
			$this->cssStyles()
		);
	}

	private function getDbCredentials() {
		$missing = [];
		if(empty($_REQUEST['dbServer'])) $missing['dbServer'] = true;
		if(empty($_REQUEST['dbUser'])) $missing['dbUser'] = true;
		if(empty($_REQUEST['dbName'])) $missing['dbName'] = true;
		if(empty($_REQUEST['dbPass'])) $_REQUEST['dbPass'] = ''; // handle the case of empty password

		if(count($missing))
			$this->askForCredentialsAndExit($missing);

		return $_REQUEST;
	}

	private function insecureConnectionWarning() {
		if(!empty($_SERVER['HTTPS']) || $_SERVER['SERVER_NAME'] == 'localhost')
			return '';

		ob_start(); ?>
		<div class="danger">
			This page is not served through a secure connection. It's dangerous to continue
			because your database credentials will be sent to the server unencrypted and
			could potentially be stolen.<br>
			Try
			<a href="<?php echo "https://{$_SERVER['SERVER_NAME']}{$_SERVER['REQUEST_URI']}"; ?>">loading the secure version of this page</a>
			if your server supports it, or continue at your own risk.
		</div>
		<?php return ob_get_clean();
	}

	private function askForCredentialsAndExit($missing = []) {
		if(count($missing) == 3) $missing = [];

		ob_start(); ?>
			<h1>Import a MySQL database to AppGini</h1>
			<p>Use this tool to import the definition of your MySQL database as an AXP project file that can be opened in <a href="https://bigprof.com/appgini/">AppGini</a>.</p>

			<?php echo $this->insecureConnectionWarning(); ?>

			<?php if(!empty($missing['error'])) { ?>
				<div class="danger"><?php echo $missing['error']; ?></div>
			<?php } ?>
 
			<form method="POST" action="<?php echo basename(__FILE__); ?>">
				<div>
					<label for="dbServer">Database server</label>
					<input name="dbServer" id="dbServer" value="localhost" autofocus="" required>
					<?php if(!empty($missing['dbServer'])) { ?>
						<div class="danger" id="dbServer-error">Invalid database server provided.</div>
					<?php } ?>
				</div>
				<div>
					<label for="dbName">Database name</label>
					<input name="dbName" id="dbName" required>
					<?php if(!empty($missing['dbName'])) { ?>
						<div class="danger" id="dbName-error">Invalid database name provided.</div>
					<?php } ?>
				</div>
				<div>
					<label for="dbUser">Database username</label>
					<input name="dbUser" id="dbUser" required>
					<?php if(!empty($missing['dbUser'])) { ?>
						<div class="danger" id="dbUser-error">Invalid database username provided.</div>
					<?php } ?>
				</div>
				<div>
					<label for="dbPass">Database password</label>
					<input type="password" name="dbPass" id="dbPass">
				</div>
				<div><button type="submit">Continue &#9654;</button></div>
			</form>

			<script>
				// hide error when related input changes
				const inputs = ['dbServer', 'dbName', 'dbUser'];
				for(let inp of inputs) {
					if(document.querySelector(`#${inp}-error`) === null) continue;
						
					document.querySelector(`#${inp}`).addEventListener('keyup', function() {
						if(this.value.length > 0)
							document.querySelector(`#${inp}-error`).classList.add('hidden');
					});
				}
			</script>

			<?php echo $this->copyrightFooter(); ?>
			<?php echo $this->cssStyles(); ?>

		<?php
		die(ob_get_clean());
	}

	private function copyrightFooter() {
		ob_start(); ?>
		<div class="copyright">
			Copyright &copy; <?php echo date('Y'); ?>,
			<a href="https://bigprof.com/appgini/m2a">BigProf Software</a>.
		</div>
		<?php return ob_get_clean();
	}

	private function cssStyles() {
		ob_start(); ?>
		<style>
			body {
				width: 100%;
				max-width: 40em;
				margin: 1.5em auto;
				font-family: sans-serif;
				line-height: 1.65em;
				padding-top: 97px;
				background-image: url("https://cdn.bigprof.com/images/appgini-flat-logo-with-text-sm.png");
				background-repeat: no-repeat;
				background-position-x: center;
				background-position-y: 27px;
			}
			.danger {
				color: red;
				background-color: #FFECEC;
				padding: 0.5em 1em;
				border: solid 1px red;
				margin-bottom: 1em;
			}
			label {
				display: block;
				font-weight: bold;
				margin-top: 1em;
			}
			input {
				width: 100%;
				font-size: 2em;
				font-family: monospace;
				color: #444;
			}
			input:focus {
				color: #000;
				border: solid 2px;
			}
			button {
				font-size: 1.25rem;
				font-weight: bold;
				margin: 1rem auto;
				display: block;
				min-width: 8rem;
				padding: 0.4rem 3rem;
			}
			ul.tables-list {
				max-height: 50vh;
				overflow-y: auto;
				border: dotted 1px #999;
				background-color: #f9f9f9;
				list-style: none;
			}
			ul.tables-list li:before {
				content: '\2714\0020';
			}
			.copyright {
				text-align: center;
				border-top: 1px solid #999;
				padding-top: 0.5rem;
				font-size: 0.8rem;
			}
			.hidden { display: none; }
			a[download] {
				font-weight: bold;
				text-align: center;
				display: block;
				text-decoration: none;
				padding: 1em;
				border: solid 1px silver;
				border-radius: 6px;
				background-color: #ddd;
			}
			a[download]:hover {
				background-color: #eee;
			}
		</style>
		<?php return ob_get_clean();
	}

	private function dataTypeAppIndex($dt) {
		$dataType = [
			'DUMMY_INDEX_BECAUSE_NO_ZERO!',
			'tinyint',
			'smallint',
			'mediumint',
			'int',
			'bigint',
			'float',
			'double',
			'decimal',
			'date',
			'datetime',
			'timestamp',
			'time',
			'year',
			'char',
			'varchar',
			'tinyblob',
			'tinytext',
			'text',
			'blob',
			'mediumblob',
			'mediumtext',
			'longblob',
			'longtext',
		];

		return array_search(strtolower($dt), $dataType) | 0;
	}

	private function getAppTableNames() {
		$tables = [];

		if(!$res = @mysqli_query($this->link, "SHOW TABLES"))
			$this->askForCredentialsAndExit([
				'error' => "Error while listing tables of the database: " . mysqli_error($this->link)
			]);

		while($row = mysqli_fetch_row($res))
			// excluded tables
			if(!preg_match('/^(membership_|appgini_)/i', $row[0]))
				$tables[] = $row[0];

		return $tables;
	}

	private function caption($name) {
		return str_replace(['_', '-'], ' ', ucfirst($name));
	}

	private function getTableFieldsXml($tn) {
		if(!$res = mysqli_query($this->link, "SHOW FIELDS FROM `$tn`"))
			return '';

		/*
		 Expected row:
		 	['Field', 'Type', 'Null', 'Key', 'Default', 'Extra']
		 */

		$axp = '';
		$i = 0;
		while($row = mysqli_fetch_assoc($res)) {
			$m = [];
			if(!preg_match('/^(\w+)(\((\d+)(,(\d+))?\))?/', $row['Type'], $m))
				continue;

			$dtUnsigned = (strpos($row['Type'], ' unsigned') ? 'True' : 'False');
			$dtZeroFill = (strpos($row['Type'], ' zerofill') ? 'True' : 'False');
			$dtType = $m[1];
			$dtLength = $m[3] ?? '';
			$dtPrecision = $m[5] ?? '';
			$dtTextArea = (stripos($dtType, 'text') !== false ? 'True' : 'False');
			
			$notNull = (($row['Null'] == '' || $row['Null'] == 'YES') ? "False" : "True");
			$primaryKey = (strtoupper($row['Key']) == 'PRI' ? "True" : "False");
			$unique = (strtoupper($row['Key']) == 'UNI' ? "True" : "False");
			$autoIncrement = (strtolower($row['Extra']) == 'auto_increment' ? "True" : "False");
			$readOnly = $autoIncrement;

			$axp .= implode('', [
				'<field>',
				'<caption><![CDATA[' . $this->caption($row['Field']) . ']]></caption>',
				'<description><![CDATA[]]></description>',
				"<name>{$row['Field']}</name>",
				'<dataType>' . $this->dataTypeAppIndex($dtType) . '</dataType>',
				"<length>$dtLength</length>",
				"<precision>$dtPrecision</precision>",
				"<autoIncrement>$autoIncrement</autoIncrement>",
				'<binary>False</binary>',
				"<notNull>$notNull</notNull>",
				"<primaryKey>$primaryKey</primaryKey>",
				"<unique>$unique</unique>",
				"<unsigned>$dtUnsigned</unsigned>",
				'<showColumnSum>False</showColumnSum>',
				"<zeroFill>$dtZeroFill</zeroFill>",
				"<default><![CDATA[{$row['Default']}]]></default>",
				'<columnWidth>150</columnWidth>',
				'<forcedWidth>False</forcedWidth>',
				'<maxLengthInTableView>0</maxLengthInTableView>',
				'<allowImageUpload>False</allowImageUpload>',
				"<readOnly>$readOnly</readOnly>",
				"<textarea>$dtTextArea</textarea>",
				'<htmlarea>False</htmlarea>',
				'<linkBehavior>0</linkBehavior>',
				'<linkDisplay>0</linkDisplay>',
				'<linkDisplayField></linkDisplayField>',
				'<checkBox>False</checkBox>',
				'<dataFormat><![CDATA[0]]></dataFormat>',
				'<filterBy></filterBy>',
				'<technicalDocumentation><![CDATA[' . implode(' ', $row) . ']]></technicalDocumentation>',
				"<index>$i</index>",
				'</field>',
			]);

			$i++;
		}

		return $axp;
	}

	private function saveAndDownloadAxp() {
		$filename = $this->dbCredentials['dbName'] . '.axp';
		if(!@file_put_contents(__DIR__ . '/' . $filename, $this->axp))
			$this->askForCredentialsAndExit([
				'error' => "Couldn't save the project file. Please chmod the folder where you uploaded this script to 777 and try again."
			]);

		?>
			<h1>MySQL2AppGini Database Import Wizard</h1>
		
			<p>Successufully imported the following tables:
				<ul class="tables-list">
					<li> <?php echo implode('</li><li> ', $this->tables); ?></li>
				</ul>
			</p>
		
			<h3><a href="<?php echo urlencode($filename); ?>" download>Download the AppGini project file</a></h3>
		
			<p>After downloading the file, you can double-click it to open in AppGini.</p>
			<div style="text-align: right;">
				<a href="<?php echo basename(__FILE__); ?>">Import another database</a>
			</div>
		<?php
		echo $this->copyrightFooter();
		echo $this->cssStyles();
	}
}